from enum import Enum

from welkin.models.base import Collection
from welkin.pagination import FormationIterator


class FormationDataType(Enum):
    CDTS = "cdts"
    ASSESSMENTS = "assessments"
    ENCOUNTERS = "encounters"
    ENCOUNTER_DISPOSITION = "encounter-disposition"
    CARE_PLAN = "goal-templates"


class Formations(Collection):
    iterator = FormationIterator

    def get(self, data_type: FormationDataType, *args, **kwargs):
        return super().get(
            f"{self._client.instance}/formations/current/{data_type.value}",
            *args,
            **kwargs,
        )



"""Client

This module provides a Client object to interface with the Welkin Health API.
"""
import logging
from json import JSONDecodeError

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter
from requests.compat import urljoin
from requests.packages.urllib3.util.retry import Retry  # type: ignore

from welkin import __version__
from welkin.authentication import WelkinAuth
from welkin.exceptions import WelkinHTTPError
from welkin.models import *
from welkin.util import clean_request_params, clean_request_payload

logger = logging.getLogger(__name__)


class Client(Session):
    """Welkin Health Client.

    Constructs a :obj:`requests.Session` for Welkin Health API requests with
    authorization, base URL, request timeouts, and request retries.

    Args:
        tenant (str): Name of the organization.
        instance (str): Name of the database inside a customer space.
        api_client (str): API client name.
        secret_key (str): API client secret key.
        timeout (int, optional): :obj:`TimeoutHTTPAdapter` timeout value. Defaults to 5.
        total (int, optional): :obj:`Retry` total value. Defaults to 5.
        backoff_factor (float, optional): :obj:`Retry` backoff_factor value.
            Defaults to 0.5.

    Usage::

        from welkin import Client

        welkin = Client(
            tenant="gh",
            instance="sb-demo",
            api_client="VBOPNRYRWJIP",
            secret_key="+}B{KGTG6#zG%P;tQm0C",
        )


        ### Patient methods
        patient = welkin.Patient(firstName="Foo", lastName="Bar").create()  # Create

        patient = welkin.Patient(id="6801d498-26f4-4aee-961b-5daffcf193c8").get()  # Read
        patients = welkin.Patients().get()  # Read all/list

        patient.update(firstName="Baz")  # Update
        patient.delete()  # Delete

        ### User methods
        user = client.User(username="bar", email="bar@foo.com").create()  # Create

        user = welkin.User(id="301b2895-cbf0-4cac-b4cf-1d082faee95c").get()  # Read
        users = welkin.Users().get()  # Read all/list
        uasers = welkin.Users().get(
            search="lightmatter", region="east-coast", seat_assigned=True, user_state="ACTIVE"
        )  # Filtered read all/list

        user.update(firstName="Baz")  # Update
        user.delete()  # Delete
    """

    def __init__(
        self,
        tenant,
        instance,
        api_client,
        secret_key,
        timeout=5,
        total=5,
        backoff_factor=0.5,
    ):
        """Welkin Health client constructor.

        Constructs a :obj:`requests.Session` for Welkin Health API requests with
        authorization, base URL, request timeouts, and request retries.

        Args:
            tenant (str): Name of the customer space.
            instance (str): Name of the database inside a customer space.
            api_client (str): API client name.
            secret_key (str): API client secret key.
            timeout (int, optional): :obj:`TimeoutHTTPAdapter` timeout value. Defaults to 5.
            total (int, optional): :obj:`Retry` total value. Defaults to 5.
            backoff_factor (float, optional): :obj:`Retry` backoff_factor value.
                Defaults to 0.5.
        """
        super().__init__()

        self.host = f"https://api.live.welkincloud.io/{tenant}/"
        self.headers.update({"User-Agent": f"python-welkin/{__version__}"})

        adapter = TimeoutHTTPAdapter(
            timeout=timeout,
            max_retries=Retry(
                total=total,
                status_forcelist=[429, 500, 502, 503, 504],
                backoff_factor=backoff_factor,
            ),
        )
        self.mount("https://", adapter)
        self.mount("http://", adapter)

        self.auth = WelkinAuth(
            tenant=tenant,
            api_client=api_client,
            secret_key=secret_key,
            token_method=self.get_token,
        )

        self.instance = instance
        self.__build_resources()

    def __build_resources(self) -> None:
        """Add each resource with a reference to this instance."""
        for k, v in globals().items():
            try:
                for base in v.__bases__:
                    if base.__name__ not in ["Collection", "Resource"]:
                        continue

                    v._client = self
                    setattr(self, k, v)

            except AttributeError:
                continue

    def prepare_request(self, request):
        if request.json:
            request.json = clean_request_payload(request.json)
        if request.params:
            request.params = clean_request_params(request.params)

        return super().prepare_request(request)

    def request(
        self,
        method: str,
        path: str,
        meta_key: str = None,
        meta_dict: dict = {},
        *args,
        **kwargs,
    ):
        """Override :obj:`Session` request method to add retries and output JSON.

        Args:
            method (str): Method for the new Request object.
            path (str): Path from host for the new Request object.

        Returns:
            dict: Response JSON
        """
        if not isinstance(path, str):
            path = "/".join((str(s) for s in path if s))
        path = path.rstrip("/")

        for _ in range(2):
            response = super().request(
                method=method, url=urljoin(self.host, path), *args, **kwargs
            )

            try:
                response.raise_for_status()
                break
            except HTTPError as exc:
                code = exc.response.status_code

                if code in [401]:
                    msg = response.json()
                    codes = ["NOT_VALID_JSON_WEB_TOKEN", "TOKEN_EXPIRED"]
                    if any(i.get("errorCode") in codes for i in msg):
                        self.auth.refresh_token()
                        continue

                raise WelkinHTTPError(exc) from exc

        try:
            json = response.json()
        except JSONDecodeError:
            if not response.content:
                return {}

            raise

        if isinstance(json, list):
            json = {
                "content": json,
                meta_key: meta_dict,
            }
        elif "rows" in json:
            json = {
                "content": json.pop("rows"),
                meta_key: meta_dict,
            }

        # Pull out the resource
        resource = json
        if "content" in json:
            resource = json.pop("content", None)
        elif "data" in json:
            resource = json.pop("data", None)

        meta = None

        # specifically with cdts the resource and metadata are both in the data dict
        if isinstance(resource, dict) and "content" in resource:
            new_resource = resource.pop("content", None)
            meta = resource.pop(meta_key, {})
            meta.update(json)
            meta.update(resource)
            resource = new_resource
        # encounter disposition formation comes as just a dictionary
        elif isinstance(resource, dict) and "formations" in path:
            resource = [resource]

        # Response metadata for pagination
        if meta_key:
            if not meta:
                meta = json.pop(meta_key, {})
            meta.update(json)
            return resource, meta

        return resource or json

    def get_token(self) -> dict:
        data = {"secret": self.auth.secret_key}
        response = self.post(f"admin/api_clients/{self.auth.api_client}", json=data)

        return response


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout, *args, **kwargs):
        """TimeoutHTTPAdapter constructor.

        Args:
            timeout (int): How many seconds to wait for the server to send data before
                giving up.
        """
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """Override :obj:`HTTPAdapter` send method to add a default timeout."""
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout

        return super().send(request, **kwargs)



import json

from requests import HTTPError


class WelkinHTTPError(HTTPError):
    """An HTTP error occurred.

    https://developers.welkinhealth.com/#errors
    """

    def __init__(self, exc):
        """WelkinHTTPError constructor.

        Parses out error information from the error object and passes on to the
        :obj:`HTTPError` constructor.

        Args:
            exc (:obj:`HTTPError`): Original exception.
        """
        request = exc.request
        response = exc.response
        msg = exc.args[0]

        try:
            msg = f"{msg}\n{json.dumps(response.json(), indent=2)}"
        except json.JSONDecodeError:
            pass

        super(HTTPError, self).__init__(msg, request=request, response=response)



"""Welkin Health API wrapper

https://developers.welkinhealth.com/

Introduction
========
When reading this API, we take a liberty and assume that you are familiar with the
following Welkin concepts, but for the sake of clarity will shortly repeat them here:

1. Tenant (Organization) - This is a customer space, dedicated to one customer. Every
customer will have its own tenant that will host customer users, apps and instances
2. Instance (Environment) - This is a separate database inside a tenant. Typical
customer will have 2-3 instances, representing customer development, testing and live
environments, as you build out your Welkin care program
3. API client - this is an auto-generated pair of key and secrets, that allows you to
access variety of API that Welkin exposes
4. Security Policies and Roles - set of rules that dictates what API your client can
access and what actions are allowed to be performed
5. Designer - Codeless editor for configuring Care Program and all the elements of that
program, including Permissions and Roles
6. Admin - Admin app that allows one to assign permissions and roles to API clients
(among other things)
7. Care - Care portal that users will be using to deliver care to patients

For better demonstration of the API, we will use the following setup:

Organization (Tenant): gh
Instance (Environment): sb-demo

Creating API Client
========
Though this is better covered in our User Guide document, we are going to repeat the
steps here, to ensure successful setup

1. Create API client in your Organization
 - Navigate to Admin -> API Clients -> Create Client
 - Copy the Client Name and Secret Key or download it.
2. Navigate to the API Client page you created
 - Configure appropriate access for the client (Instance Access, Roles, Security
 Policies)

Reminder: Security Policies and Roles are defined in the Designer and assigned in the
Admin

For this example we will assume Client Name is VBOPNRYRWJIP and Secret Key is
+}B{KGTG6#zG%P;tQm0C
"""
from welkin.__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)
from welkin.client import Client

__all__ = ["Client"]



from welkin.models.assessment import (
    Assessment,
    AssessmentRecord,
    AssessmentRecordAnswers,
    AssessmentRecords,
    Assessments,
)
from welkin.models.calendar import CalendarEvent, CalendarEvents, Schedules
from welkin.models.care_plan import CarePlan, CarePlanOverview
from welkin.models.cdt import CDT, CDTs
from welkin.models.chat import Chat, Chats, SearchChats
from welkin.models.encounter import Disposition, Encounter, Encounters
from welkin.models.formation import Formations
from welkin.models.patient import Patient, Patients
from welkin.models.user import User, Users

# NOTE: If a class isn't imported here and added to __all__, it will not be callable
# from a `Client` instance. Also, for legibility, keep this list alphabetical.
__all__ = [
    "Assessment",
    "AssessmentRecord",
    "AssessmentRecordAnswers",
    "AssessmentRecords",
    "Assessments",
    "CalendarEvent",
    "CalendarEvents",
    "CarePlan",
    "CarePlanOverview",
    "CDT",
    "CDTs",
    "Chat",
    "Chats",
    "SearchChats",
    "Disposition",
    "Encounter",
    "Encounters",
    "Patient",
    "Patients",
    "Schedules",
    "User",
    "Users",
    "Formations",
]



from sys import modules


def find_patient_id_in_parents(model_instance):
    if isinstance(
        model_instance,
        getattr(modules["welkin.models"], "Patient"),
    ):
        return model_instance.id
    elif hasattr(model_instance, "patientId"):
        return model_instance.patientId
    elif model_instance._parent:
        return find_patient_id_in_parents(model_instance._parent)
    else:
        raise Exception(
            f"Cannot find patient id. Model._parent chain ends in {model_instance}"
        )


class EncounterSubResource:
    """Utility class for subresources of Encounters to get patient and encounter ids"""

    @property
    def patient_id(self):
        if isinstance(
            self._parent._parent,
            getattr(modules["welkin.models"], "Patient"),
        ):
            return self._parent._parent.id

        if hasattr(self._parent, "patientId"):
            return self._parent.patientId

        # this is the related_data = True case on encounters
        return self._parent.encounter.patientId

    def get_patient_encounter_id(self, patient_id, encounter_id):
        """Helper to retrieve the necessary patient and encounter Ids"""
        if not patient_id:
            patient_id = self.patient_id

        if not encounter_id:
            encounter_id = self._parent.id

        return patient_id, encounter_id


def patient_id(func):
    """Wrapper for getting the patient id on patient subresources"""

    def wrapper(cls, *args, **kwargs):
        if "patient_id" not in kwargs:
            try:
                kwargs["patient_id"] = cls._parent.id
            except AttributeError:
                raise TypeError(
                    f"{func.__name__} is missing 1 required positional argument: "
                    "'patient_id"
                ) from None

        return func(cls, *args, **kwargs)

    return wrapper



import pytest

from welkin.models.base import Resource
from welkin.models.formation import FormationDataType, Formations


@pytest.mark.vcr()
def test_formation_get(client, vcr_cassette):
    for data_type in FormationDataType:
        formation_info = client.Formations().get(data_type)
        assert isinstance(formation_info, Formations)
        if len(formation_info) > 0:
            assert isinstance(formation_info[0], Resource)

    assert len(vcr_cassette) == len(FormationDataType)



import dbm

from requests import Request

from welkin.authentication import DB_PATH


def auth_class(client):
    auth = client.auth
    if auth.tenant != "gh":
        auth.tenant = "test_tenant"
    auth.token_method = lambda: {"token": "API_TOKEN"}

    return auth


def test_token_get(client):
    auth = auth_class(client)
    token = auth.token

    assert token is not None
    assert dbm.whichdb(DB_PATH) is not None


def test_token_refresh(client):
    auth = auth_class(client)

    token = f"{auth.token}_1"
    auth.refresh_token()

    assert token != auth.token


def test_auth_call(client):
    req = Request("GET", "https://foo.com/bar")
    prepped = req.prepare()

    auth = auth_class(client)
    auth(prepped)

    assert "Authorization" in prepped.headers


def test_auth_token_call(client):
    req = Request("GET", "https://foo.com/bar/api_clients")
    prepped = req.prepare()

    auth = auth_class(client)
    auth(prepped)

    assert "Authorization" not in prepped.headers



import pytest

from welkin.models import CarePlan


@pytest.mark.vcr()
def test_care_plan_create(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")

    overview_message = "This is a message for the overview."

    care_plan = patient.CarePlan(overview=overview_message)

    created_cp = care_plan.create()

    assert isinstance(created_cp, CarePlan)
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_care_plan_create_passing_patient_id(client, vcr_cassette):
    patient_id = "331678c1-0ec9-4e4a-a947-8147a1d067a5"
    overview_message = "This is a message for the overview."

    care_plan = client.CarePlan(overview=overview_message).create(patient_id=patient_id)

    assert isinstance(care_plan, CarePlan)
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_care_plan_read(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")
    care_plan = patient.CarePlan().get()

    assert isinstance(care_plan, CarePlan)
    assert care_plan.id
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_care_plan_read_passing_patient_id(client, vcr_cassette):
    patient_id = "331678c1-0ec9-4e4a-a947-8147a1d067a5"
    care_plan = client.CarePlan().get(patient_id=patient_id)

    assert isinstance(care_plan, CarePlan)
    assert care_plan.id
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_care_plan_update(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")
    care_plan = patient.CarePlan().get()
    overview = care_plan.patientOverview["overview"]
    new_overview = "A new overview message 123"
    care_plan.CarePlanOverview(**{"overview": new_overview}).update()
    care_plan = patient.CarePlan().get()

    assert care_plan.patientOverview["overview"] != overview
    assert care_plan.patientOverview["overview"] == new_overview
    assert len(vcr_cassette) == 3


@pytest.mark.vcr()
def test_care_plan_update_passing_patient_id(client, vcr_cassette):
    patient_id = "173a8adf-92e8-4832-8900-027c71b0d768"
    care_plan = client.CarePlan().get(patient_id=patient_id)
    overview = care_plan.patientOverview["overview"]
    new_overview = "Why not have a new overview message?"
    care_plan = client.CarePlanOverview(**{"overview": new_overview}).update(
        patient_id=patient_id
    )
    care_plan = client.CarePlan().get(patient_id=patient_id)

    assert care_plan.patientOverview["overview"] != overview
    assert care_plan.patientOverview["overview"] == new_overview
    assert len(vcr_cassette) == 3



from welkin.models.assessment import AssessmentRecord, AssessmentRecords
from welkin.models.base import Collection, Resource
from welkin.models.care_plan import CarePlan
from welkin.models.cdt import CDT, CDTs
from welkin.models.chat import Chat, Chats, SearchChats
from welkin.models.encounter import Encounter, Encounters
from welkin.pagination import PageableIterator


class Patient(Resource):
    subresources = [
        AssessmentRecord,
        AssessmentRecords,
        CarePlan,
        CDT,
        CDTs,
        Chat,
        Chats,
        SearchChats,
        Encounter,
        Encounters,
    ]

    def create(self):
        return super().post(f"{self._client.instance}/patients")

    def get(self):
        return super().get(f"{self._client.instance}/patients/{self.id}")

    def update(self, **kwargs):
        return super().patch(f"{self._client.instance}/patients/{self.id}", kwargs)

    def delete(self):
        return super().delete(f"{self._client.instance}/patients/{self.id}")

    def __str__(self):
        return f"{self.firstName} {self.lastName}"


class Patients(Collection):
    resource = Patient
    iterator = PageableIterator

    def get(self, filter={}, *args, **kwargs):
        # TODO: Add sort and query arguments.
        return super().post(
            f"{self._client.instance}/by-filter/patients", json=filter, *args, **kwargs
        )



from datetime import datetime, timedelta, timezone

import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.calendar import CalendarEvent, CalendarEvents, Schedule, Schedules

UTC = timezone.utc


@pytest.mark.vcr()
def test_calendar_event_create(client, vcr_cassette):
    start = datetime.now(tz=UTC) + timedelta(hours=6)
    end = start + timedelta(hours=1)

    event = client.CalendarEvent(
        startDateTime=start,
        endDateTime=end,
        hostId="d9925247-7505-464a-b42c-25ac71408494",
        participants=[
            {
                "participantId": "d9925247-7505-464a-b42c-25ac71408494",
                "participantRole": "psm",
            }
        ],
    ).create()

    assert isinstance(event, CalendarEvent)
    assert hasattr(event, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_calendar_event_read(client, vcr_cassette):
    event_id = "9386a88a-467e-4144-a7df-1af12d6d5aaf"
    calendar = client.CalendarEvent(id=event_id).get()

    assert isinstance(calendar, CalendarEvent)
    assert calendar.id == event_id
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_calendar_event_read_all(client, vcr_cassette):
    from_date = datetime(2022, 4, 1, 16, 38, 20, 641000, tzinfo=UTC)
    to_date = datetime(2022, 6, 30, 16, 38, 20, 641000, tzinfo=UTC)

    events = client.CalendarEvents().get(from_date, to_date, include_cancelled=True)

    assert isinstance(events, CalendarEvents)
    assert isinstance(events[0], CalendarEvent)

    if len(events) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_calendar_event_update(client, vcr_cassette):
    event = client.CalendarEvent(id="f7fc881d-28af-4fe1-b2ff-20e1ca33982f").get()
    title = event.eventTitle

    event.update(eventTitle="New event title")

    assert event.eventTitle != title
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_calendar_event_delete(client, vcr_cassette):
    event = client.CalendarEvent(id="f7fc881d-28af-4fe1-b2ff-20e1ca33982f").get()
    event.delete()

    if event.startDateTime <= datetime.now(tz=UTC).isoformat():
        event.get()
        assert event.eventStatus == "CANCELLED"

    else:
        with pytest.raises(WelkinHTTPError) as excinfo:
            event.get()

        assert excinfo.value.response.status_code == 404

    assert len(vcr_cassette) == 3


@pytest.mark.vcr()
def test_schedule_read_all(client, vcr_cassette):
    from_date = datetime(2022, 4, 1, 16, 38, 20, 641000, tzinfo=UTC)
    to_date = datetime(2022, 6, 30, 16, 38, 20, 641000, tzinfo=UTC)

    schedules = client.Schedules().get(
        ids=["d9925247-7505-464a-b42c-25ac71408494"],
        from_date=from_date,
        to_date=to_date,
        include_cancelled=True,
    )

    assert isinstance(schedules, Schedules)
    assert isinstance(schedules[0], Schedule)

    assert len(vcr_cassette) == 1



from datetime import datetime

from welkin.models.base import Collection, Resource
from welkin.pagination import PageNumberIterator


class CDT(Resource):
    def create(self):
        return super().post(
            f"{self._client.instance}/patients/{self._parent.id}/cdts/{self.cdtName}"
        )

    def get(self):
        return super().get(
            f"{self._client.instance}/patients/{self._parent.id}/cdts/{self.cdtName}/{self.id}"
        )

    def update(self, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{self._parent.id}/cdts/{self.cdtName}/{self.id}",
            kwargs,
        )

    def delete(self):
        return super().delete(
            f"{self._client.instance}/patients/{self._parent.id}/cdts/{self.cdtName}/{self.id}"
        )


class CDTs(Collection):
    resource = CDT
    iterator = PageNumberIterator

    def get(
        self,
        patient_id: str = None,
        cdt_name: str = None,
        fields: list = None,
        filters: dict = None,
        date_start: datetime = None,
        date_end: datetime = None,
        sort: str = None,
        *args,
        **kwargs,
    ):

        root = ""
        if patient_id:
            root = f"patients/{patient_id}"
        else:
            root = f"patients/{self._parent.id}"

        cdt_path = ""
        if cdt_name:
            cdt_path = f"cdts/{cdt_name}"

        path = f"{self._client.instance}/{root}/{cdt_path}"

        params = {
            "fields": fields,
            "filters": filters,
            "sort": sort,
            "dateStart": date_start,
            "dateEnd": date_end,
        }

        return super().get(path, params=params, *args, **kwargs)

    def update(
        self, patient_id: str = None, cdt_name: str = None, body: dict = None, **kwargs
    ):
        root = ""
        if patient_id:
            root = f"patients/{patient_id}"
        else:
            root = f"patients/{self._parent.id}"

        cdt_path = ""
        if cdt_name:
            cdt_path = f"cdts/{cdt_name}"

        path = f"{self._client.instance}/{root}/{cdt_path}"

        return super().patch(path, json=body, **kwargs)



from datetime import datetime, timedelta, timezone

import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.assessment import Assessment, Assessments
from welkin.models.encounter import Disposition, Encounter, Encounters
from welkin.models.user import User

UTC = timezone.utc


@pytest.mark.vcr()
def test_encounter_create(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")
    start = datetime.now(tz=UTC) + timedelta(hours=6)
    end = start + timedelta(hours=1)

    data = {
        "title": "new enc",
        "description": "new encounter",
        "templateName": "etmp-initial-consultation",
        "currentScheduledAppointment": {
            "eventType": "ENCOUNTER",
            "eventTitle": "NEW ENC",
            "startDateTime": start,
            "endDateTime": end,
            "hostId": "c08af975-8afc-49cb-84ac-7189b727148c",
            "participants": [
                {
                    "participantId": "c08af975-8afc-49cb-84ac-7189b727148c",
                    "participantRole": "psm",
                },
                {"participantId": patient.id, "participantRole": "patient"},
            ],
        },
    }

    encounter = patient.Encounter(**data)

    created_enc = encounter.create()

    assert isinstance(created_enc, Encounter)
    assert hasattr(created_enc, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_all_encounters_patient_read(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")
    encounters = patient.Encounters().get()

    assert isinstance(encounters, Encounters)
    assert isinstance(encounters[0], Encounter)

    if len(encounters) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_all_encounters_user_read(client, vcr_cassette):
    user = client.User(id="c08af975-8afc-49cb-84ac-7189b727148c")
    encounters = user.Encounters().get(with_care_team=False)

    assert isinstance(encounters, Encounters)
    assert isinstance(encounters[0], Encounter)

    if len(encounters) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_encounter_patient_read(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="2a8ac491-19bf-4812-9197-dffc2d42cfcf").get()

    assert isinstance(encounter, Encounter)
    assert encounter.id == "2a8ac491-19bf-4812-9197-dffc2d42cfcf"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_patient_read_related_data(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get(
        related_data=True
    )

    assert isinstance(encounter, Encounter)
    assert isinstance(encounter.userRelatedToCalendarEvent, User)
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_update(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get()
    notes = encounter.jsonBody["notes"]

    encounter.update(notes="a newer note")

    assert encounter.jsonBody["notes"] != notes
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_encounter_delete(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="2a8ac491-19bf-4812-9197-dffc2d42cfcf").get()

    encounter.delete()

    with pytest.raises(WelkinHTTPError) as excinfo:
        encounter.get()

        assert excinfo.value.response.status_code == 404

    assert len(vcr_cassette) == 3


@pytest.mark.vcr()
def test_encounter_assessments(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get(
        related_data=True
    )

    assert isinstance(encounter, Encounter)
    assert isinstance(encounter.assessmentLinks, Assessments)
    assert isinstance(encounter.assessmentLinks[0], Assessment)
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_assessments_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get()
    assessments = encounter.Assessments().get()

    assert isinstance(assessments, Assessments)
    assert isinstance(assessments[0], Assessment)
    assert len(vcr_cassette) == 2


# these tests show all of the ways one could get an Encounter Assessment
@pytest.mark.vcr()
def test_encounter_assessment_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6")
    assessment = encounter.Assessment(id="7cf6baa2-14d5-4d3a-9416-0ddd729644b8").get()

    assert isinstance(assessment, Assessment)
    assert assessment.id == "7cf6baa2-14d5-4d3a-9416-0ddd729644b8"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_get_assessment_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get()
    assessment = encounter.Assessment(id="10c6481b-f25b-49cf-9761-33aeced25f46").get()

    assert isinstance(assessment, Assessment)
    assert assessment.id == "10c6481b-f25b-49cf-9761-33aeced25f46"
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_assessment_get(client, vcr_cassette):
    assessment = client.Assessment(id="10c6481b-f25b-49cf-9761-33aeced25f46").get(
        patient_id="371dd15c-cedc-4425-a394-d666c8d3fc01",
        encounter_id="22c26a65-161f-42c3-bb0f-a976dac8afe6",
    )
    assert isinstance(assessment, Assessment)
    assert assessment.id == "10c6481b-f25b-49cf-9761-33aeced25f46"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_related_data_assessment_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6").get(
        related_data=True
    )
    assessment = encounter.assessmentLinks[-1].get()

    assert isinstance(assessment, Assessment)
    assert assessment.id == "10c6481b-f25b-49cf-9761-33aeced25f46"
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_encounter_assessment_create(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6")
    assessment = encounter.Assessment(assessmentName="asm-hello-form2").create()

    assert isinstance(assessment, Assessment)
    assert hasattr(assessment, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_assessment_update(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6")

    assessment = encounter.Assessment(id="4d2be3ba-2be3-4171-a8bf-de68cd8e5a42").get()
    assessment_record_id = assessment.jsonBody["assessmentRecordId"]

    assessment.update(assessmentRecordId="e18a2759-5089-44e7-9d31-cf942476ffab")

    assert isinstance(assessment, Assessment)
    assert assessment.jsonBody["assessmentRecordId"] != assessment_record_id
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_encounter_assessment_delete(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    encounter = patient.Encounter(id="22c26a65-161f-42c3-bb0f-a976dac8afe6")
    assessment = encounter.Assessment(id="4d2be3ba-2be3-4171-a8bf-de68cd8e5a42")

    assessment.delete()
    with pytest.raises(WelkinHTTPError) as excinfo:
        assessment.get()

        assert excinfo.value.response.status_code == 404

    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_encounter_disposition_get_nested(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")
    encounter = patient.Encounter(id="d6f4b66e-1be6-403a-ae47-1bbcee264c5e").get(
        related_data=True
    )
    disposition = encounter.disposition

    assert isinstance(disposition, Disposition)
    assert hasattr(disposition, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_disposition_get(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")
    encounter = patient.Encounter(id="d6f4b66e-1be6-403a-ae47-1bbcee264c5e")
    disposition = encounter.Disposition().get()

    assert isinstance(disposition, Disposition)
    assert disposition.id == "8cd87974-b1cb-4a6f-8873-71d5d188c4aa"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_disposition_get_patient_id_encounter_id(client, vcr_cassette):
    disposition = client.Disposition().get(
        patient_id="173a8adf-92e8-4832-8900-027c71b0d768",
        encounter_id="d6f4b66e-1be6-403a-ae47-1bbcee264c5e",
    )
    assert isinstance(disposition, Disposition)
    assert disposition.id == "8cd87974-b1cb-4a6f-8873-71d5d188c4aa"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_encounter_disposition_update(client, vcr_cassette):
    patient = client.Patient(id="173a8adf-92e8-4832-8900-027c71b0d768")
    encounter = patient.Encounter(id="d6f4b66e-1be6-403a-ae47-1bbcee264c5e")
    disposition = encounter.Disposition(id="8cd87974-b1cb-4a6f-8873-71d5d188c4aa").get()

    review = disposition.jsonBody["uicedf-review"]

    disposition.update(**{"uicedf-review": "great health"})

    assert disposition.jsonBody["uicedf-review"] != review
    assert len(vcr_cassette) == 2



from datetime import date

import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.patient import Patient, Patients


@pytest.mark.vcr()
def test_patient_create(client, vcr_cassette):
    patient = client.Patient(firstName="Foo", lastName="Bar").create()

    assert isinstance(patient, Patient)
    assert hasattr(patient, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_patient_create_birthdate(client, vcr_cassette):
    patient = client.Patient(
        firstName="happy", lastName="borf", birthDate=date.today()
    ).create()

    assert isinstance(patient, Patient)
    assert hasattr(patient, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_patient_read(client, vcr_cassette):
    patient_id = "092ae416-1c0d-4e14-be22-9cc8dafacdbd"
    patient = client.Patient(id=patient_id).get()

    assert isinstance(patient, Patient)
    assert patient.id == patient_id
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_patient_read_all(client, vcr_cassette):
    patients = client.Patients().get()

    assert isinstance(patients, Patients)
    assert isinstance(patients[0], Patient)

    if len(patients) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_patient_update(client, vcr_cassette):
    patient = client.Patient(id="092ae416-1c0d-4e14-be22-9cc8dafacdbd").get()
    name = patient.firstName

    patient.update(firstName="Baz")

    assert patient.firstName != name
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_patient_delete(client, vcr_cassette):
    patient = client.Patient(id="092ae416-1c0d-4e14-be22-9cc8dafacdbd")

    with pytest.raises(WelkinHTTPError) as excinfo:
        patient.delete()

    assert excinfo.value.response.status_code == 403

    assert len(vcr_cassette) == 1



import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.user import User, Users


@pytest.mark.vcr()
def test_user_create(client, vcr_cassette):
    user = client.User(username="bar", email="bar@foo.com").create()

    assert isinstance(user, User)
    assert hasattr(user, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_user_read(client, vcr_cassette):
    user = client.User(id="658f3b36-a12a-4bce-aad6-1e4930948b7d")

    assert isinstance(user, User)
    assert len(user) == 1

    user.get()

    assert len(user) > 1
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_user_read_all(client, vcr_cassette):
    users = client.Users().get()

    assert isinstance(users, Users)
    assert isinstance(users[0], User)

    if len(users) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_user_update(client, vcr_cassette):
    user = client.User(id="92cc5811-ff71-4101-915d-a419383db168").get()
    name = user.firstName

    user.update(firstName="Baz")

    assert user.firstName != name
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_user_delete(client, vcr_cassette):
    user = client.User(id="658f3b36-a12a-4bce-aad6-1e4930948b7d")

    with pytest.raises(WelkinHTTPError) as excinfo:
        user.delete()

    assert excinfo.value.response.status_code == 403

    assert len(vcr_cassette) == 1



import pytest

from welkin.models.chat import Chat, Chats, ChatSearchResult, SearchChats


@pytest.mark.vcr()
def test_create_chat(client, vcr_cassette):
    patient_id = "17450e44-c2c8-46c4-9486-0d9bfa16d3aa"
    message = "Foo Baz."
    patient = client.Patient(id=patient_id)
    chat = patient.Chat(message=message).create()

    assert isinstance(chat, Chat)
    assert chat.message == message
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_read_chat(client, vcr_cassette):
    patient_id = "17450e44-c2c8-46c4-9486-0d9bfa16d3aa"
    patient = client.Patient(id=patient_id)
    chats = patient.Chats().get()

    assert isinstance(chats, Chats)
    if len(chats) > 0:
        assert isinstance(chats[0], Chat)
        assert chats[0].message
        assert str(chats[0]) == "PATIENT Foo Baz."
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_search_chat(client, vcr_cassette):
    patient_id = "17450e44-c2c8-46c4-9486-0d9bfa16d3aa"
    query = "Test"
    patient = client.Patient(id=patient_id)
    chats = patient.SearchChats().get(query=query)

    assert isinstance(chats, SearchChats)
    if len(chats) > 0:
        assert isinstance(chats[0], ChatSearchResult)
        assert query.lower() in chats[0].message["message"].lower()

    if len(chats) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"



from datetime import date, datetime, timezone

# NOTE: `clean_request_payload` and `clean_request_params` are intentionally DRY
# violations. The code may be the same, but they represent different knowledge.
# the same goes for `clean_request_payload` and `clean_json_list`


def clean_request_payload(payload: dict) -> dict:
    for k, v in payload.items():
        if isinstance(v, datetime):
            payload[k] = clean_datetime(v)
        elif isinstance(v, date):
            payload[k] = clean_date(v)
        elif isinstance(v, dict):
            payload[k] = clean_request_payload(v)
        elif isinstance(v, list):
            payload[k] = clean_json_list(v)

    return payload


def clean_json_list(data: list) -> list:
    for ind, item in enumerate(data):
        if isinstance(item, datetime):
            data[ind] = clean_datetime(item)
        elif isinstance(item, date):
            data[ind] = clean_date(item)
        elif isinstance(item, dict):
            data[ind] = clean_request_payload(item)
        elif isinstance(item, list):
            data[ind] = clean_json_list(item)

    return data


def clean_request_params(params: dict) -> dict:
    for k, v in params.items():
        if isinstance(v, datetime):
            params[k] = clean_datetime(v)
        elif isinstance(v, date):
            params[k] = clean_date(v)
        elif isinstance(v, list):
            params[k] = ",".join(v)

    return params


def clean_date(date: date) -> str:
    dt = datetime(date.year, date.month, date.day, 0, 0, 0)

    return clean_datetime(dt)


def clean_datetime(dt: datetime) -> str:
    return (
        dt.astimezone(tz=timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )



from welkin.models.base import Resource
from welkin.models.util import find_patient_id_in_parents


class CarePlanOverview(Resource):
    def update(self, patient_id: str = None):
        if not patient_id:
            # self._parent -> CarePlan
            # CarePlan._parent -> Patient
            patient_id = find_patient_id_in_parents(self)
        return super().put(
            f"{self._client.instance}/patients/{patient_id}/care-plan/overview"
        )


class CarePlan(Resource):
    subresources = [CarePlanOverview]

    def create(self, patient_id: str = None):
        if not patient_id:
            # self._parent -> Patient
            patient_id = find_patient_id_in_parents(self)
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/care-plan/overview"
        )

    def get(self, patient_id: str = None):
        if not patient_id:
            # self._parent -> Patient
            patient_id = find_patient_id_in_parents(self)
        return super().get(f"{self._client.instance}/patients/{patient_id}/care-plan")



"""Package version information.
"""

__title__ = "welkin"
__description__ = "Python Welkin Health API Wrapper."
__url__ = "https://github.com/Lightmatter/welkin-health"
__version__ = "0.0.6"
__author__ = "Sam Morgan"
__author_email__ = "sam@lightmatter.com"
__license__ = "GNU GPL 3.0"
__copyright__ = "2022 Lightmatter"



import inspect

from welkin.models.base import Collection
from welkin.pagination import PageIterator


def test_collection_pageable(client):
    for v in vars(client).values():
        try:
            if not issubclass(v, Collection):
                continue
        except TypeError:
            continue

        assert hasattr(v, "get"), f"{v} has no get method"
        assert issubclass(v.iterator, PageIterator)

        args = inspect.getfullargspec(v.get)
        assert args.varargs is not None, f"{v}.get must accept variable args"
        assert args.varkw is not None, f"{v}.get must accept variable kwargs"



from sys import modules

from welkin.models.base import Collection, Resource
from welkin.models.util import EncounterSubResource
from welkin.pagination import PageableIterator


class Assessment(Resource, EncounterSubResource):
    def create(self, patient_id: str = None, encounter_id: str = None):
        patient_id, encounter_id = self.get_patient_encounter_id(
            patient_id, encounter_id
        )
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}"
            "/assessments"
        )

    def get(self, patient_id: str = None, encounter_id: str = None):
        patient_id, encounter_id = self.get_patient_encounter_id(
            patient_id, encounter_id
        )
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}"
        )

    def update(self, patient_id: str = None, encounter_id: str = None, **kwargs):
        patient_id, encounter_id = self.get_patient_encounter_id(
            patient_id, encounter_id
        )
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}",
            kwargs,
        )

    def delete(self, patient_id: str = None, encounter_id: str = None):
        patient_id, encounter_id = self.get_patient_encounter_id(
            patient_id, encounter_id
        )
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}"
        )


class Assessments(Collection):
    resource = Assessment
    iterator = PageableIterator

    def get(self, patient_id: str = None, encounter_id: str = None, *args, **kwargs):

        root = f"{self._client.instance}/patients/"
        if self._parent:
            encounter_id = self._parent.id
            if isinstance(
                self._parent._parent, getattr(modules["welkin.models"], "Patient")
            ):
                patient_id = self._parent._parent.id
            elif hasattr(self._parent, "patientId"):
                patient_id = self._parent.patientId
            else:
                # this is the related_data = True case on encounters
                patient_id = self._parent.encounter.patientId

        path = f"{patient_id}/encounters/{encounter_id}/assessments"

        return super().get(f"{root}{path}", *args, **kwargs)


class AssessmentRecordAnswers(Resource):
    def update(self, patient_id: str = None, assessment_record_id: str = None):
        if not assessment_record_id:
            assessment_record_id = self._parent.id

        if not patient_id:
            patient_id = self._parent.get_patient_id(patient_id)

        return super().put(
            f"{self._client.instance}/patients/{patient_id}/"
            f"assessment-records/{assessment_record_id}/answers"
        )


def patient_id(func):
    def wrapper(cls, *args, **kwargs):
        if "patient_id" not in kwargs:
            try:
                kwargs["patient_id"] = cls._parent.id
            except AttributeError:
                raise TypeError(
                    f"{func.__name__} is missing 1 required positional argument: "
                    "'patient_id"
                ) from None

        return func(cls, *args, **kwargs)

    return wrapper


class AssessmentRecord(Resource):
    subresources = [AssessmentRecordAnswers]

    @patient_id
    def create(self, patient_id: str = None):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/assessment-records"
        )

    @patient_id
    def get(self, patient_id: str = None):
        return super().get(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}"
        )

    @patient_id
    def update(self, patient_id: str = None):
        return super().put(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}"
        )

    @patient_id
    def delete(self, patient_id: str = None):
        return super().delete(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}",
        )

    def get_patient_id(self, patient_id):
        return patient_id if patient_id else self._parent.id


class AssessmentRecords(Collection):
    resource = AssessmentRecord
    iterator = PageableIterator

    @patient_id
    def get(self, patient_id: str = None, **kwargs):
        path = f"{self._client.instance}/patients/{patient_id}/assessment-records"

        return super().get(path, **kwargs)



import json
import os
import uuid

import pytest

from welkin import Client


def pytest_collection_modifyitems(items):
    # Ensure auth tests execute first, otherwise all other tests will fail.
    items.sort(key=lambda x: True if "authentication" not in x.nodeid else False)


def redact(field_name, extra=""):
    parts = [field_name, extra, str(uuid.uuid4())]

    return "_".join(i for i in parts if i)


HEADER_BLACKLIST = [("Authorization", redact("API_TOKEN"))]
POST_DATA_BLACKLIST = [("secret", redact("API_TOKEN"))]
REQUEST_BLACKLIST = ["secret"]
RESPONSE_BLACKLIST = [
    "token",
    "createdBy",
    "createdByName",
    "updatedBy",
    "updatedByName",
]
CLIENT_INIT = dict(
    tenant=os.environ["WELKIN_TENANT"],
    instance=os.environ["WELKIN_INSTANCE"],
    api_client=os.environ["WELKIN_API_CLIENT"],
    secret_key=os.environ["WELKIN_SECRET"],
)


@pytest.fixture(scope="module")
def client():
    """Get an authenticated Welkin API client."""
    return Client(**CLIENT_INIT)


@pytest.fixture(scope="module")
def vcr(vcr):
    vcr.filter_headers = HEADER_BLACKLIST
    vcr.filter_post_data_parameters = POST_DATA_BLACKLIST
    vcr.before_record_request = scrub_request(CLIENT_INIT)
    vcr.before_record_response = scrub_response(RESPONSE_BLACKLIST)

    return vcr


def scrub_request(blacklist, replacement="REDACTED"):
    def before_record_request(request):
        # request.body = filter_body(request.body, blacklist, replacement)
        if "api_clients" in request.path:
            return None
        uri_comps = request.uri.split("/")
        for k, v in blacklist.items():
            try:
                ind = uri_comps.index(v)
                uri_comps[ind] = f"{k}_{replacement}"
            except ValueError:
                continue

        request.uri = "/".join(uri_comps)
        return request

    return before_record_request


def scrub_response(blacklist, replacement="REDACTED"):
    def before_record_response(response):
        response["body"]["string"] = filter_body(
            response["body"]["string"], blacklist, replacement
        )

        return response

    return before_record_response


def filter_body(body, blacklist, replacement):
    if not body:
        return body
    object_hook = body_hook(blacklist, replacement)
    body_json = json.loads(body.decode(), object_hook=object_hook)

    return json.dumps(body_json).encode()


def body_hook(blacklist, replacement):
    def hook(dct):
        for k in dct:
            if k in blacklist:
                dct[k] = redact(k, replacement)

        return dct

    return hook



from pathlib import Path

from setuptools import find_packages, setup

root = Path(__file__).parent

with (root / "requirements.txt").open() as f:
    requires = f.read().splitlines()

about = {}
with (root / "welkin" / "__version__.py").open() as f:
    exec(f.read(), about)

with (root / "README.md").open() as f:
    readme = f.read()

setup(
    author_email=about["__author_email__"],
    author=about["__author__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description=about["__description__"],
    install_requires=requires,
    license=about["__license__"],
    long_description_content_type="text/markdown",
    long_description=readme,
    name=about["__title__"],
    packages=find_packages(),
    project_urls={
        "Documentation": about["__url__"],
        "Source": "https://github.com/Lightmatter/welkin-health",
        "Tracker": "https://github.com/Lightmatter/welkin-health/issues",
    },
    python_requires="~=3.6",
    url=about["__url__"],
    version=about["__version__"],
)



import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.cdt import CDT, CDTs


@pytest.mark.vcr()
def test_cdt_create(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")

    cdt = patient.CDT(cdtName="cdt-self-care", **{"cdtf-alone-time": "Yes"}).create()

    assert isinstance(cdt, CDT)
    assert hasattr(cdt, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_all_cdts_patient_read(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")
    cdts = patient.CDTs().get(cdt_name="cdt-self-care")

    assert isinstance(cdts, CDTs)
    assert isinstance(cdts[0], CDT)

    if len(cdts) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr()
def test_cdt_patient_read(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")

    cdt = patient.CDT(
        cdtName="cdt-self-care", id="20a1cd2c-dd70-4589-afc4-fb99adaa25ad"
    ).get()

    assert isinstance(cdt, CDT)
    assert cdt.id == "20a1cd2c-dd70-4589-afc4-fb99adaa25ad"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_cdt_update(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")
    cdt = patient.CDT(
        cdtName="cdt-self-care", id="20a1cd2c-dd70-4589-afc4-fb99adaa25ad"
    ).get()
    alone_time_response = cdt.jsonBody["cdtf-alone-time"]
    cdt.update(**{"cdtf-alone-time": "No"})
    assert cdt.jsonBody["cdtf-alone-time"] != alone_time_response
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_cdt_bulk_update(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")
    cdt_1 = patient.CDT(
        cdtName="cdt-self-care", id="20a1cd2c-dd70-4589-afc4-fb99adaa25ad"
    ).get()
    cdt_2 = patient.CDT(
        cdtName="cdt-self-care", id="28f59572-efec-4ac1-bbfe-660b1abbd9e5"
    ).get()

    resp_1 = "Yes" if cdt_1.jsonBody["cdtf-alone-time"] == "No" else "Yes"
    resp_2 = "Yes" if cdt_2.jsonBody["cdtf-alone-time"] == "No" else "Yes"

    data = {
        "rows": [
            {
                "id": "20a1cd2c-dd70-4589-afc4-fb99adaa25ad",
                "jsonBody": {
                    "cdtf-alone-time": resp_1,
                },
            },
            {
                "id": "28f59572-efec-4ac1-bbfe-660b1abbd9e5",
                "jsonBody": {"cdtf-alone-time": resp_2},
            },
        ]
    }

    updated = patient.CDTs().update(cdt_name="cdt-self-care", body=data)
    assert updated[0].jsonBody["cdtf-alone-time"] == resp_1
    assert updated[1].jsonBody["cdtf-alone-time"] == resp_2
    assert len(vcr_cassette) == 3


@pytest.mark.vcr()
def test_cdt_delete(client, vcr_cassette):
    patient = client.Patient(id="3ff322fd-b504-4784-9798-0c7c8a4fdfb8")
    cdt = patient.CDT(
        cdtName="cdt-self-care", id="20a1cd2c-dd70-4589-afc4-fb99adaa25ad"
    ).get()

    cdt.delete()

    with pytest.raises(WelkinHTTPError) as excinfo:
        cdt.get()

        assert excinfo.value.response.status_code == 404

    assert len(vcr_cassette) == 3



import pytest

from welkin.exceptions import WelkinHTTPError
from welkin.models.assessment import AssessmentRecord, AssessmentRecords


@pytest.mark.vcr()
def test_assessment_record_answers_update(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assmt_record = patient.AssessmentRecord(
        id="c8764b19-ffa3-406a-b446-c971036a7a1d"
    ).get()
    hi_answer = assmt_record.answers["cdt-hello__cdtf-hi"]
    assmt_record.AssessmentRecordAnswers(**{"cdt-hello__cdtf-hi": "abc123def"}).update()
    assmt_record = patient.AssessmentRecord(
        id="c8764b19-ffa3-406a-b446-c971036a7a1d"
    ).get()

    assert assmt_record.answers["cdt-hello__cdtf-hi"] == "abc123def"
    assert hi_answer != assmt_record.answers["cdt-hello__cdtf-hi"]
    assert len(vcr_cassette) == 3


@pytest.mark.vcr()
def test_assessment_record_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assmt_record = patient.AssessmentRecord(
        id="c8764b19-ffa3-406a-b446-c971036a7a1d"
    ).get()

    assert isinstance(assmt_record, AssessmentRecord)
    assert assmt_record.id == "c8764b19-ffa3-406a-b446-c971036a7a1d"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_assessment_record_get_w_patient_id(client, vcr_cassette):
    assmt_record = client.AssessmentRecord(
        id="c8764b19-ffa3-406a-b446-c971036a7a1d"
    ).get(patient_id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assert isinstance(assmt_record, AssessmentRecord)
    assert assmt_record.id == "c8764b19-ffa3-406a-b446-c971036a7a1d"
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_assessment_record_update(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assmt_record = patient.AssessmentRecord(
        id="e18a2759-5089-44e7-9d31-cf942476ffab"
    ).get()
    status = assmt_record.status

    assmt_record.status = "COMPLETED"
    assmt_record.update()

    assert assmt_record.status != status
    assert assmt_record.status == "COMPLETED"
    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_assessment_record_create(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")
    assmt_record = patient.AssessmentRecord(assessmentName="asm-hello-form2").create()

    assert isinstance(assmt_record, AssessmentRecord)
    assert hasattr(assmt_record, "id")
    assert len(vcr_cassette) == 1


@pytest.mark.vcr()
def test_assessment_record_delete(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assmt_record = patient.AssessmentRecord(id="4f44ef74-9c82-4174-902b-1f9efebdebb7")

    assmt_record.delete()

    with pytest.raises(WelkinHTTPError) as excinfo:
        assmt_record.get()

        assert excinfo.value.response.status_code == 404

    assert len(vcr_cassette) == 2


@pytest.mark.vcr()
def test_assessment_records_get(client, vcr_cassette):
    patient = client.Patient(id="371dd15c-cedc-4425-a394-d666c8d3fc01")

    assmt_records = patient.AssessmentRecords().get()

    assert isinstance(assmt_records, AssessmentRecords)
    assert isinstance(assmt_records[0], AssessmentRecord)
    if len(assmt_records) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"


@pytest.mark.vcr(vcr_cassette_name="test_assessment_records_get_patient_id.yaml")
def test_assessment_records_get_patient_id(client, vcr_cassette):
    assmt_records = client.AssessmentRecords().get(
        patient_id="371dd15c-cedc-4425-a394-d666c8d3fc01"
    )

    assert isinstance(assmt_records, AssessmentRecords)
    assert isinstance(assmt_records[0], AssessmentRecord)
    if len(assmt_records) > 20:
        assert len(vcr_cassette) > 1, "Pagination was expected"
    else:
        assert len(vcr_cassette) == 1, "Unexpected pagination"



class PageIterator:
    def __init__(self, collection, resource, method, size=20, *args, **kwargs):
        self.collection = collection
        self.resource = resource
        self.method = method
        self.size = size
        self.meta_key = None
        self.meta_dict = {"totalPages": 1, "number": 0, "last": True}

        if size != 20:
            kwargs.setdefault("params", {}).update(size=size)

        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        self._resources = []
        self.last = False

        return self

    def __next__(self):
        if self.resources:
            return self.resources.pop(0)

        if not self.last:
            self._pre_request()

            self.resources, meta = self.method(
                self.resource,
                meta_key=self.meta_key,
                meta_dict=self.meta_dict,
                *self.args,
                **self.kwargs,
            )

            self._post_request(meta)

            return next(self)

        raise StopIteration

    def _pre_request(self):
        """
        Function to execute before making the next request, e.g. update paging params
        """
        self.kwargs.setdefault("params", {})

    def _post_request(self, meta):
        """
        Function to execute after making the next request, e.g. updating page tracking
        """
        self.last = True

    @property
    def resources(self):
        return self._resources

    @resources.setter
    def resources(self, value):
        self._resources = [self.collection.resource(v) for v in value]
        self.collection.extend(self.resources)


class PageableIterator(PageIterator):
    """
    Most common paging class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_key = "pageable"

    def __iter__(self):
        self.page = 0
        return super().__iter__()

    def _pre_request(self):
        self.kwargs.setdefault("params", {}).update(page=self.page)

    def _post_request(self, meta):
        self.page = meta["number"] + 1
        self.last = meta.get("last")


class FormationIterator(PageIterator):
    """
    Specifically for paginating formations responses
    Similar to PageableIterator but includes special behavior for
    single item formations, e.g. encounter disposition
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_key = "pageable"

    def __iter__(self):
        self.page = 0
        return super().__iter__()

    def _pre_request(self):
        self.kwargs.setdefault("params", {}).update(page=self.page)

    def _post_request(self, meta):
        self.page = meta.get("number", 0) + 1
        self.last = meta.get("last", True)


class PageNumberIterator(PageableIterator):
    """
    PageableIterator with a different key used for the page count
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_dict.pop("number")
        self.meta_dict["pageNumber"] = 0

    def _post_request(self, meta):
        self.page = meta["pageNumber"] + 1
        self.last = meta.get("last")


class MetaInfoIterator(PageIterator):
    """
    Functionally identical to PageableIterator with various renamed keys
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_key = "metaInfo"
        self.meta_dict.pop("number")
        self.meta_dict["page"] = 0

    def __iter__(self):
        self.page = 0
        return super().__iter__()

    def _pre_request(self):
        self.kwargs.setdefault("params", {}).update(page=self.page)

    def _post_request(self, meta):
        self.page = meta["page"] + 1
        self.last = meta.get("lastPage")


class MetaIterator(PageIterator):
    """
    Paging class for token based paging
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_key = "meta"
        self.meta_dict = {
            "nextPageToken": None,
            "prevPageToken": None,
            "pageSize": self.size,
            "found": False,
        }
        self.kwargs.setdefault("params", {}).update(pageSize=self.size)

        if "page_token" in kwargs:
            self.page_token = kwargs["page_token"]

    def __iter__(self):
        self.page_token = None
        return super().__iter__()

    def _pre_request(self):
        if self.page_token:
            self.kwargs["params"] = {"pageToken": self.page_token}

    def _post_request(self, meta):
        self.page_token = meta["nextPageToken"]
        self.last = not meta.get("nextPageToken")
